@@ -62,6 +62,7 @@ import (
 )
 
 func (s *Site) Taxonomies() page.TaxonomyList {
+	s.checkReady()
 	s.init.taxonomies.Do(context.Background())
 	return s.taxonomies
 }
@@ -204,6 +205,7 @@ type siteRenderingContext struct {
 }
 
 func (s *Site) Menus() navigation.Menus {
+	s.checkReady()
 	s.init.menus.Do(context.Background())
 	return s.menus
 }
@@ -372,19 +374,33 @@ func (s *Site) watching() bool {
 type whatChanged struct {
 	mu sync.Mutex
 
-	contentChanged bool
-	identitySet    identity.Identities
+	needsPagesAssembly bool
+	identitySet        identity.Identities
 }
 
 func (w *whatChanged) Add(ids ...identity.Identity) {
 	w.mu.Lock()
 	defer w.mu.Unlock()
 
+	if w.identitySet == nil {
+		w.identitySet = make(identity.Identities)
+	}
+
 	for _, id := range ids {
 		w.identitySet[id] = true
 	}
 }
 
+func (w *whatChanged) Clear() {
+	w.mu.Lock()
+	defer w.mu.Unlock()
+	w.clear()
+}
+
+func (w *whatChanged) clear() {
+	w.identitySet = identity.Identities{}
+}
+
 func (w *whatChanged) Changes() []identity.Identity {
 	if w == nil || w.identitySet == nil {
 		return nil
@@ -392,6 +408,14 @@ func (w *whatChanged) Changes() []identity.Identity {
 	return w.identitySet.AsSlice()
 }
 
+func (w *whatChanged) Drain() []identity.Identity {
+	w.mu.Lock()
+	defer w.mu.Unlock()
+	ids := w.identitySet.AsSlice()
+	w.clear()
+	return ids
+}
+
 // RegisterMediaTypes will register the Site's media types in the mime
 // package, so it will behave correctly with Hugo's built-in server.
 func (s *Site) RegisterMediaTypes() {
@@ -786,6 +810,7 @@ func (s *Site) errorCollator(results <-chan error, errs chan<- error) {
 // as possible for existing sites. Most sites will use {{ .Site.GetPage "section" "my/section" }},
 // i.e. 2 arguments, so we test for that.
 func (s *Site) GetPage(ref ...string) (page.Page, error) {
+	s.checkReady()
 	p, err := s.s.getPageForRefs(ref...)
 
 	if p == nil {
